@@ -397,7 +397,7 @@ MoreLink.prototype = {
 
         let text = new Clutter.Text({ font_name: "Sans 12px",
                                       color: BRIGHT_TEXT_COLOR,
-                                      text: _("Browse") });
+                                      text: _("More") });
         this.actor.append(text, Big.BoxPackFlags.NONE);
 
         this.actor.connect('button-press-event', Lang.bind(this, function (b, e) {
@@ -506,6 +506,13 @@ SectionHeader.prototype = {
             this.backLink.actor.hide();
     },
 
+    setMoreLinkVisible : function(visible) {
+        if (visible)
+            this.moreLink.actor.show();
+        else
+            this.moreLink.actor.hide();
+    },
+
     setCountText : function(countText) {
         if (countText == "") {
             this.countText.hide();
@@ -613,6 +620,7 @@ Dash.prototype = {
         /***** Search *****/
 
         this._searchActive = false;
+        this._searchPending = false;
         this._searchEntry = new SearchEntry();
         this.searchArea.append(this._searchEntry.actor, Big.BoxPackFlags.EXPAND);
 
@@ -620,7 +628,9 @@ Dash.prototype = {
         this._searchEntry.entry.connect('text-changed', Lang.bind(this, function (se, prop) {
             let text = this._searchEntry.getText();
             text = text.replace(/^\s+/g, "").replace(/\s+$/g, "")
+            let searchPreviouslyActive = this._searchActive;
             this._searchActive = text != '';
+            this._searchPending = this._searchActive && !searchPreviouslyActive;
             this._updateDashActors();
             if (!this._searchActive) {
                 if (this._searchTimeoutId > 0) {
@@ -779,8 +789,8 @@ Dash.prototype = {
 
         this._docsSection = new Section(_("RECENT DOCUMENTS"));
 
-        let docDisplay = new DocDisplay.DashDocDisplay();
-        this._docsSection.content.append(docDisplay.actor, Big.BoxPackFlags.EXPAND);
+        this._docDisplay = new DocDisplay.DashDocDisplay();
+        this._docsSection.content.append(this._docDisplay.actor, Big.BoxPackFlags.EXPAND);
 
         this._moreDocsPane = null;
         this._docsSection.header.moreLink.connect('activated', Lang.bind(this, function (link) {
@@ -792,6 +802,12 @@ Dash.prototype = {
            }
         }));
 
+        this._docDisplay.connect('changed', Lang.bind(this, function () {
+            this._docsSection.header.setMoreLinkVisible(
+                this._docDisplay.actor.get_children().length > 0);
+        }));
+        this._docDisplay.emit('changed');
+
         this.sectionArea.append(this._docsSection.actor, Big.BoxPackFlags.EXPAND);
 
         /***** Search Results *****/
@@ -871,17 +887,25 @@ Dash.prototype = {
     },
 
     _updateDashActors: function() {
-        if (!this._searchActive && this._searchResultsSection.actor.visible) {
+        if (this._searchPending) {
+            this._searchResultsSection.actor.show();
+            // We initially hide all sections when we start a search. When the search timeout
+            // first runs, the sections that have matching results are shown. As the search
+            // is refined, only the sections that have matching results will be shown.
+            for (let i = 0; i < this._searchSections.length; i++) {
+                let section = this._searchSections[i];
+                section.header.actor.hide();
+                section.resultArea.actor.hide();
+            }
+            this._appsSection.actor.hide();
+            this._placesSection.actor.hide();
+            this._docsSection.actor.hide();
+        } else if (!this._searchActive) {
             this._showAllSearchSections();
             this._searchResultsSection.actor.hide();
             this._appsSection.actor.show();
             this._placesSection.actor.show();
             this._docsSection.actor.show();
-        } else if (this._searchActive && !this._searchResultsSection.actor.visible) {
-            this._searchResultsSection.actor.show();
-            this._appsSection.actor.hide();
-            this._placesSection.actor.hide();
-            this._docsSection.actor.hide();
         }
     },
 
