@@ -102,7 +102,7 @@ DocDisplayItem.prototype = {
     _resetTimeDisplay: function(currentSecs) {
         let lastSecs = this._docInfo.timestamp;
         let timeDelta = currentSecs - lastSecs;
-        let [text, nextUpdate] = Shell.Global.get().format_time_relative_pretty(timeDelta);
+        let [text, nextUpdate] = global.format_time_relative_pretty(timeDelta);
         this._timeoutTime = currentSecs + nextUpdate;
         this._setDescriptionText(text);
     }
@@ -120,8 +120,6 @@ DocDisplay.prototype = {
 
     _init : function() {
         GenericDisplay.GenericDisplay.prototype._init.call(this);
-        let me = this;
-
         // We keep a single timeout callback for updating last visited times
         // for all the items in the display. This avoids creating individual
         // callbacks for each item in the display. So proper time updates
@@ -132,14 +130,14 @@ DocDisplay.prototype = {
 
         this._docManager = DocInfo.getDocManager();
         this._docsStale = true;
-        this._docManager.connect('changed', function(mgr, userData) {
-            me._docsStale = true;
+        this._docManager.connect('changed', Lang.bind(this, function(mgr, userData) {
+            this._docsStale = true;
             // Changes in local recent files should not happen when we are in the Overview mode,
             // but redisplaying right away is cool when we use Zephyr.
             // Also, we might be displaying remote documents, like Google Docs, in the future
             // which might be edited by someone else.
-            me._redisplay(false);
-        });
+            this._redisplay(GenericDisplay.RedisplayFlags.NONE);
+        }));
 
         this.connect('destroy', Lang.bind(this, function (o) {
             if (this._updateTimeoutId > 0)
@@ -152,9 +150,10 @@ DocDisplay.prototype = {
     // Gets the list of recent items from the recent items manager.
     _refreshCache : function() {
         if (!this._docsStale)
-            return;
+            return true;
         this._allItems = this._docManager.getItems();
         this._docsStale = false;
+        return false;
     },
 
     // Sets the list of the displayed items based on how recently they were last visited.
@@ -173,21 +172,24 @@ DocDisplay.prototype = {
         // them once when they are returned by this._recentManager.get_items() to avoid having to do 
         // this sorting each time, but the sorting seems to be very fast anyway, so there is no need
         // to introduce an additional class variable.
-        this._matchedItems = [];
+        this._matchedItems = {};
+        this._matchedItemKeys = [];
         let docIdsToRemove = [];
         for (docId in this._allItems) {
             // this._allItems[docId].exists() checks if the resource still exists
-            if (this._allItems[docId].exists()) 
-                this._matchedItems.push(docId);
-            else 
+            if (this._allItems[docId].exists()) {
+                this._matchedItems[docId] = 1;
+                this._matchedItemKeys.push(docId);
+            } else {
                 docIdsToRemove.push(docId);
+            }
         }
 
         for (docId in docIdsToRemove) {
             delete this._allItems[docId];
         }
 
-        this._matchedItems.sort(Lang.bind(this, function (a,b) { return this._compareItems(a,b); }));
+        this._matchedItemKeys.sort(Lang.bind(this, this._compareItems));
     },
 
     // Compares items associated with the item ids based on how recently the items
@@ -300,7 +302,7 @@ DashDocDisplayItem.prototype = {
     shellWorkspaceLaunch: function () {
         this._info.launch();
     }
-}
+};
 
 /**
  * Class used to display two column recent documents in the dash
@@ -434,5 +436,9 @@ DashDocDisplay.prototype = {
             let display = new DashDocDisplayItem(docInfo);
             this.actor.add_actor(display.actor);
         }
+        this.emit('changed');
     }
-}
+};
+
+Signals.addSignalMethods(DashDocDisplay.prototype);
+
