@@ -1,6 +1,7 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 const Clutter = imports.gi.Clutter;
+const DBus = imports.dbus;
 const Gdk = imports.gi.Gdk;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
@@ -9,14 +10,16 @@ const Mainloop = imports.mainloop;
 const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
+const St = imports.gi.St;
 
 const Chrome = imports.ui.chrome;
+const Environment = imports.ui.environment;
 const Overview = imports.ui.overview;
 const Panel = imports.ui.panel;
 const RunDialog = imports.ui.runDialog;
 const LookingGlass = imports.ui.lookingGlass;
+const ShellDBus = imports.ui.shellDBus;
 const Sidebar = imports.ui.sidebar;
-const Tweener = imports.ui.tweener;
 const WindowManager = imports.ui.windowManager;
 
 const DEFAULT_BACKGROUND_COLOR = new Clutter.Color();
@@ -30,6 +33,7 @@ let runDialog = null;
 let lookingGlass = null;
 let wm = null;
 let recorder = null;
+let shellDBusService = null;
 let modalCount = 0;
 let modalActorFocusStack = [];
 
@@ -42,8 +46,14 @@ function start() {
     Gio.DesktopAppInfo.set_desktop_env("GNOME");
 
     global.grab_dbus_service();
+    shellDBusService = new ShellDBus.GnomeShell();
+    // Force a connection now; dbus.js will do this internally
+    // if we use its name acquisition stuff but we aren't right
+    // now; to do so we'd need to convert from its async calls
+    // back into sync ones.
+    DBus.session.flush();
 
-    Tweener.init();
+    Environment.init();
 
     // Ensure ShellAppMonitor is initialized; this will
     // also initialize ShellAppSystem first.  ShellAppSystem
@@ -66,10 +76,24 @@ function start() {
     for (let i = 0; i < children.length; i++)
         children[i].destroy();
 
+    let themeContext = St.ThemeContext.get_for_stage (global.stage);
+    let stylesheetPath = global.datadir + "/theme/gnome-shell.css";
+    let theme = new St.Theme ({ application_stylesheet: stylesheetPath });
+    themeContext.set_theme (theme);
+
     global.connect('panel-run-dialog', function(panel) {
         // Make sure not more than one run dialog is shown.
         getRunDialog().open();
     });
+    let shellwm = global.window_manager;
+    shellwm.takeover_keybinding("panel_main_menu");
+    shellwm.connect("keybinding::panel_main_menu", function () {
+        overview.toggle();
+    });
+    shellwm.takeover_keybinding("panel_run_dialog");
+    shellwm.connect("keybinding::panel_run_dialog", function () {
+       getRunDialog().open();
+    });
 
     overview = new Overview.Overview();
     chrome = new Chrome.Chrome();
@@ -103,7 +127,9 @@ function start() {
 }
 
 function _relayout() {
-    panel.actor.set_size(global.screen_width, Panel.PANEL_HEIGHT);
+    let primary = global.get_primary_monitor();
+    panel.actor.set_position(primary.x, primary.y);
+    panel.actor.set_size(primary.width, Panel.PANEL_HEIGHT);
     overview.relayout();
 }
 
@@ -175,7 +201,7 @@ function _globalKeyPressHandler(actor, event) {
                 overview.hide();
 
             return true;
-        } else if (symbol == Clutter.F2 && (event.get_state() & Clutter.ModifierType.MOD1_MASK)) {
+        } else if (symbol == Clutter.F2 && (Shell.get_event_state(event) & Clutter.ModifierType.MOD1_MASK)) {
             getRunDialog().open();
         }
     }
@@ -205,9 +231,18 @@ function _findModal(actor) {
  * Next, record the current Clutter keyboard focus on a stack.  If the modal stack
  * returns to this actor, reset the focus to the actor which was focused
  * at the time pushModal() was invoked.
+ *
+ * Returns: true iff we successfully acquired a grab or already had one
  */
 function pushModal(actor) {
-    let timestamp = global.screen.get_display().get_current_time();
+    if (modalCount == 0) {
+        if (!global.begin_modal(currentTime())) {
+            log("pushModal: invocation of begin_modal failed");
+            return false;
+        }
+    }
+
+    global.set_stage_input_mode(Shell.StageInputMode.FULLSCREEN);
 
     modalCount += 1;
     actor.connect('destroy', function() {
@@ -225,14 +260,7 @@ function pushModal(actor) {
     }
     modalActorFocusStack.push([actor, curFocus]);
 
-    if (modalCount > 1)
-        return;
-
-    if (!global.begin_modal(timestamp)) {
-        log("pushModal: invocation of begin_modal failed");
-        return;
-    }
-    global.set_stage_input_mode(Shell.StageInputMode.FULLSCREEN);
+    return true;
 }
 
 /**
@@ -244,8 +272,6 @@ function pushModal(actor) {
  * previous focus at the time when pushModal() was invoked.
  */
 function popModal(actor) {
-    let timestamp = global.screen.get_display().get_current_time();
-
     modalCount -= 1;
     let focusIndex = _findModal(actor);
     if (focusIndex >= 0) {
@@ -263,7 +289,7 @@ function popModal(actor) {
     if (modalCount > 0)
         return;
 
-    global.end_modal(timestamp);
+    global.end_modal(currentTime());
     global.set_stage_input_mode(Shell.StageInputMode.NORMAL);
 }
 
@@ -283,15 +309,62 @@ function getRunDialog() {
 }
 
 function createAppLaunchContext() {
-    let screen = global.screen;
-    let display = screen.get_display();
-
     let context = new Gdk.AppLaunchContext();
-    context.set_timestamp(display.get_current_time());
+    context.set_timestamp(currentTime());
 
     // Make sure that the app is opened on the current workspace even if
     // the user switches before it starts
-    context.set_desktop(screen.get_active_workspace_index());
+    context.set_desktop(global.screen.get_active_workspace_index());
 
     return context;
 }
+
+/**
+ * currentTime:
+ *
+ * Gets the current X server time from the current Clutter, Gdk, or X
+ * event. If called from outside an event handler, this may return
+ * %Clutter.CURRENT_TIME (aka 0), or it may return a slightly
+ * out-of-date timestamp.
+ */
+function currentTime() {
+    // meta_display_get_current_time() will return the correct time
+    // when handling an X or Gdk event, but will return CurrentTime
+    // from some Clutter event callbacks.
+    //
+    // clutter_get_current_event_time() will return the correct time
+    // from a Clutter event callback, but may return an out-of-date
+    // timestamp if called at other times.
+    //
+    // So we try meta_display_get_current_time() first, since we
+    // can recognize a "wrong" answer from that, and then fall back
+    // to clutter_get_current_event_time().
+
+    let time = global.screen.get_display().get_current_time();
+    if (time != Clutter.CURRENT_TIME)
+        return time;
+
+    return Clutter.get_current_event_time();
+}
+
+/**
+ * activateWindow:
+ * @window: the Meta.Window to activate
+ * @time: (optional) current event time
+ *
+ * Activates @window, switching to its workspace first if necessary
+ */
+function activateWindow(window, time) {
+    let activeWorkspaceNum = global.screen.get_active_workspace_index();
+    let windowWorkspaceNum = window.get_workspace().index();
+
+    if (!time)
+        time = currentTime();
+
+    if (windowWorkspaceNum != activeWorkspaceNum) {
+        let workspace = global.screen.get_workspace_by_index(windowWorkspaceNum);
+        workspace.activate_with_focus(window, time);
+    } else {
+        window.activate(time);
+    }
+}
