@@ -12,6 +12,7 @@ const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 
 const DND = imports.ui.dnd;
+const Lightbox = imports.ui.lightbox;
 const Main = imports.ui.main;
 const Overview = imports.ui.overview;
 const Panel = imports.ui.panel;
@@ -25,8 +26,6 @@ const WINDOWCLONE_TITLE_COLOR = new Clutter.Color();
 WINDOWCLONE_TITLE_COLOR.from_pixel(0xffffffff);
 const FRAME_COLOR = new Clutter.Color();
 FRAME_COLOR.from_pixel(0xffffffff);
-const LIGHTBOX_COLOR = new Clutter.Color();
-LIGHTBOX_COLOR.from_pixel(0x00000044);
 
 const SCROLL_SCALE_AMOUNT = 100 / 5;
 
@@ -45,7 +44,8 @@ const POSITIONS = {
         4: [[0.25, 0.25, 0.47],   [0.75, 0.25, 0.47], [0.75, 0.75, 0.47], [0.25, 0.75, 0.47]],
         5: [[0.165, 0.25, 0.32], [0.495, 0.25, 0.32], [0.825, 0.25, 0.32], [0.25, 0.75, 0.32], [0.75, 0.75, 0.32]]
 };
-
+// Used in _orderWindowsPermutations, 5! = 120 which is probably the highest we can go
+const POSITIONING_PERMUTATIONS_MAX = 5;
 
 function _interpolate(start, end, step) {
     return start + (end - start) * step;
@@ -111,6 +111,8 @@ WindowClone.prototype = {
         this.origX = realWindow.x;
         this.origY = realWindow.y;
 
+        this._stackAbove = null;
+
         this._title = null;
 
         this.actor.connect('button-release-event',
@@ -129,6 +131,8 @@ WindowClone.prototype = {
         this._draggable.connect('drag-begin', Lang.bind(this, this._onDragBegin));
         this._draggable.connect('drag-end', Lang.bind(this, this._onDragEnd));
         this._inDrag = false;
+
+        this._zooming = false;
     },
 
     setVisibleWithChrome: function(visible) {
@@ -143,6 +147,14 @@ WindowClone.prototype = {
         }
     },
 
+    setStackAbove: function (actor) {
+        this._stackAbove = actor;
+        if (this._inDrag || this._zooming)
+            // We'll fix up the stack after the drag/zooming
+            return;
+        this.actor.raise(this._stackAbove);
+    },
+
     destroy: function () {
         this.actor.destroy();
         if (this._title)
@@ -157,7 +169,6 @@ WindowClone.prototype = {
 
         this._havePointer = true;
 
-        actor.raise_top();
         this._updateTitle();
     },
 
@@ -206,33 +217,19 @@ WindowClone.prototype = {
     },
 
     _zoomStart : function () {
-        this._zoomOverlay = new Clutter.Rectangle({ reactive: true,
-                                                    color: LIGHTBOX_COLOR,
-                                                    border_width: 0,
-                                                    x: 0,
-                                                    y: 0,
-                                                    width: global.screen_width,
-                                                    height: global.screen_height,
-                                                    opacity: 0 });
-        this._zoomOverlay.show();
-        global.stage.add_actor(this._zoomOverlay);
-        Tweener.addTween(this._zoomOverlay,
-                         { opacity: 255,
-                           time: ZOOM_OVERLAY_FADE_TIME,
-                           transition: "easeOutQuad"
-                         });
+        this._zooming = true;
+
+        this._zoomLightbox = new Lightbox.Lightbox(global.stage);
 
         this._zoomLocalOrig  = new ScaledPoint(this.actor.x, this.actor.y, this.actor.scale_x, this.actor.scale_y);
         this._zoomGlobalOrig = new ScaledPoint();
         let parent = this._origParent = this.actor.get_parent();
-        [width, height] = this.actor.get_transformed_size();
+        let [width, height] = this.actor.get_transformed_size();
         this._zoomGlobalOrig.setPosition.apply(this._zoomGlobalOrig, this.actor.get_transformed_position());
         this._zoomGlobalOrig.setScale(width / this.actor.width, height / this.actor.height);
 
-        this._zoomOverlay.raise_top();
-        this._zoomOverlay.show();
-
         this.actor.reparent(global.stage);
+        this._zoomLightbox.highlight(this.actor);
 
         [this.actor.x, this.actor.y]             = this._zoomGlobalOrig.getPosition();
         [this.actor.scale_x, this.actor.scale_y] = this._zoomGlobalOrig.getScale();
@@ -248,14 +245,17 @@ WindowClone.prototype = {
     },
 
     _zoomEnd : function () {
+        this._zooming = false;
+
         this.actor.reparent(this._origParent);
+        this.actor.raise(this._stackAbove);
 
         [this.actor.x, this.actor.y]             = this._zoomLocalOrig.getPosition();
         [this.actor.scale_x, this.actor.scale_y] = this._zoomLocalOrig.getScale();
 
         this._adjustTitle();
 
-        this._zoomOverlay.destroy();
+        this._zoomLightbox.destroy();
         Main.overview.disconnect(this._hideEventId);
 
         this._zoomLocalPosition  = undefined;
@@ -263,8 +263,8 @@ WindowClone.prototype = {
         this._zoomGlobalPosition = undefined;
         this._zoomGlobalScale    = undefined;
         this._zoomTargetPosition = undefined;
-        this._zoomStep       = undefined;
-        this._zoomOverlay    = undefined;
+        this._zoomStep           = undefined;
+        this._zoomLightbox       = undefined;
     },
 
     _onButtonRelease : function (actor, event) {
@@ -288,6 +288,12 @@ WindowClone.prototype = {
         // mysteriously present
         this._havePointer = false;
 
+        // We may not have a parent if DnD completed successfully, in
+        // which case our clone will shortly be destroyed and replaced
+        // with a new one on the target workspace.
+        if (this.actor.get_parent() != null)
+            this.actor.raise(this._stackAbove);
+
         this.emit('drag-end');
     },
 
@@ -299,7 +305,6 @@ WindowClone.prototype = {
     // Called by Tweener
     onAnimationComplete : function () {
         this._updateTitle();
-        this.actor.raise(this.stackAbove);
     },
 
     _createTitle : function () {
@@ -431,14 +436,6 @@ Workspace.prototype = {
         this.actor.height = global.screen_height;
         this.scale = 1.0;
 
-        this._lightbox = new Clutter.Rectangle({ color: LIGHTBOX_COLOR });
-        this.actor.connect('notify::allocation', Lang.bind(this, function () {
-            let [width, height] = this.actor.get_size();
-            this._lightbox.set_size(width, height);
-        }));
-        this.actor.add_actor(this._lightbox);
-        this._lightbox.hide();
-
         let windows = global.get_windows().filter(this._isMyWindow, this);
 
         // Find the desktop window
@@ -573,10 +570,10 @@ Workspace.prototype = {
      */
     setLightboxMode: function (showLightbox) {
         if (showLightbox) {
-            this.setHighlightWindow(null);
-            this._lightbox.show();
+            this._lightbox = new Lightbox.Lightbox(this.actor);
         } else {
-            this._lightbox.hide();
+            this._lightbox.destroy();
+            this._lightbox = null;
         }
     },
 
@@ -587,13 +584,12 @@ Workspace.prototype = {
      * Draw the user's attention to the given window @metaWindow.
      */
     setHighlightWindow: function (metaWindow) {
-        for (let i = 0; i < this._windows.length; i++) {
-            this._windows[i].actor.lower(this._lightbox);
-        }
+        let actor;
         if (metaWindow != null) {
             let clone = this.lookupCloneForMetaWindow(metaWindow);
-            clone.actor.raise(this._lightbox);
+            actor = clone.actor;
         }
+        this._lightbox.highlight(actor);
     },
 
     _adjustRemoveButton : function() {
@@ -643,12 +639,43 @@ Workspace.prototype = {
                              this._desktop.actor.height + 2 * FRAME_SIZE / this.actor.scale_y);
     },
 
+    _isCloneVisible: function(clone) {
+        return this._showOnlyWindows == null || (clone.metaWindow in this._showOnlyWindows);
+    },
+
+    /**
+     * _getVisibleClones:
+     *
+     * Returns a list WindowClone objects where the clone isn't filtered
+     * out by any application filter.  The clone for the desktop is excluded.
+     * The returned array will always be newly allocated; it is not in any
+     * defined order, and thus it's convenient to call .sort() with your
+     * choice of sorting function.
+     */
+    _getVisibleClones: function() {
+        let visible = [];
+
+        for (let i = 1; i < this._windows.length; i++) {
+            let clone = this._windows[i];
+
+            if (!this._isCloneVisible(clone))
+                continue;
+
+            visible.push(clone);
+        }
+        return visible;
+    },
+
+    _getVisibleWindows: function() {
+        return this._getVisibleClones().map(function (clone) { return clone.metaWindow; });
+    },
+
     _resetCloneVisibility: function () {
         for (let i = 1; i < this._windows.length; i++) {
             let clone = this._windows[i];
             let icon = this._windowIcons[i];
 
-            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows)) {
+            if (!this._isCloneVisible(clone)) {
                 clone.setVisibleWithChrome(false);
                 icon.hide();
             } else {
@@ -657,6 +684,220 @@ Workspace.prototype = {
         }
     },
 
+    // Only use this for n <= 20 say
+    _factorial: function(n) {
+        let result = 1;
+        for (let i = 2; i <= n; i++)
+            result *= i;
+        return result;
+    },
+
+    /**
+     * _permutation:
+     * @permutationIndex: An integer from [0, list.length!)
+     * @list: (inout): Array of objects to permute; will be modified in place
+     *
+     * Given an integer between 0 and length of array, re-order the array in-place
+     * into a permutation denoted by the index.
+     */
+    _permutation: function(permutationIndex, list) {
+        for (let j = 2; j <= list.length; j++) {
+            let firstIndex = (permutationIndex % j);
+            let secondIndex = j - 1;
+            // Swap
+            let tmp = list[firstIndex];
+            list[firstIndex] = list[secondIndex];
+            list[secondIndex] = tmp;
+            permutationIndex = Math.floor(permutationIndex / j);
+        }
+    },
+
+    /**
+     * _forEachPermutations:
+     * @list: Array
+     * @func: Function which takes a single array argument
+     *
+     * Call @func with each permutation of @list as an argument.
+     */
+    _forEachPermutations: function(list, func) {
+        let nCombinations = this._factorial(list.length);
+        for (let i = 0; i < nCombinations; i++) {
+            let listCopy = list.concat();
+            this._permutation(i, listCopy);
+            func(listCopy);
+        }
+    },
+
+    /**
+     * _computeWindowMotion:
+     * @metaWindow: A #MetaWindow
+     * @slot: An element of #POSITIONS
+     * @slotGeometry: Layout of @slot
+     *
+     * Returns a number corresponding to how much perceived motion
+     * would be involved in moving the window to the given slot.
+     * Currently this is the square of the distance between the
+     * centers.
+     */
+    _computeWindowMotion: function (metaWindow, slot, slotGeometry) {
+        let rect = new Meta.Rectangle();
+        metaWindow.get_outer_rect(rect);
+
+        let [slotX, slotY, slotWidth, slotHeight] = slotGeometry;
+        let distanceSquared;
+        let xDelta, yDelta;
+
+        xDelta = (rect.x + rect.width / 2) - (slotX + slotWidth / 2);
+        yDelta = (rect.y + rect.height / 2) - (slotY + slotHeight / 2);
+        distanceSquared = xDelta * xDelta + yDelta * yDelta;
+
+        return distanceSquared;
+    },
+
+    /**
+     * _orderWindowsPermutations:
+     *
+     * Iterate over all permutations of the windows, and determine the
+     * permutation which has the least total motion.
+     */
+    _orderWindowsPermutations: function (windows, slots, slotGeometries) {
+        let minimumMotionPermutation = null;
+        let minimumMotion = -1;
+        let permIndex = 0;
+        this._forEachPermutations(windows, Lang.bind(this, function (permutation) {
+            let motion = 0;
+            for (let i = 0; i < permutation.length; i++) {
+                let metaWindow = permutation[i];
+                let slot = slots[i];
+                let slotAbsGeometry = slotGeometries[i];
+
+                let delta = this._computeWindowMotion(metaWindow, slot, slotAbsGeometry);
+
+                motion += delta;
+            }
+
+            if (minimumMotionPermutation == null || motion < minimumMotion) {
+                minimumMotionPermutation = permutation;
+                minimumMotion = motion;
+            }
+            permIndex++;
+        }));
+        return minimumMotionPermutation;
+    },
+
+    /**
+     * _orderWindowsGreedy:
+     *
+     * Iterate over available slots in order, placing into each one the window
+     * we find with the smallest motion to that slot.
+     */
+    _orderWindowsGreedy: function(windows, slots, slotGeometries) {
+        let result = [];
+        let slotIndex = 0;
+        // Copy since we mutate below
+        let windowCopy = windows.concat();
+        for (let i = 0; i < slots.length; i++) {
+            let slot = slots[i];
+            let slotGeometry = slotGeometries[i];
+            let minimumMotionIndex = -1;
+            let minimumMotion = -1;
+            for (let j = 0; j < windowCopy.length; j++) {
+                let metaWindow = windowCopy[j];
+                let delta = this._computeWindowMotion(metaWindow, slot, slotGeometry);
+                if (minimumMotionIndex == -1 || delta < minimumMotion) {
+                    minimumMotionIndex = j;
+                    minimumMotion = delta;
+                }
+            }
+            result.push(windowCopy[minimumMotionIndex]);
+            windowCopy.splice(minimumMotionIndex, 1);
+        }
+        return result;
+    },
+
+    /**
+     * _orderWindowsByMotionAndStartup:
+     * @windows: Array of #MetaWindow
+     * @slots: Array of slots
+     *
+     * Returns a copy of @windows, ordered in such a way that they require least motion
+     * to move to the final screen coordinates of @slots.  Ties are broken in a stable
+     * fashion by the order in which the windows were created.
+     */
+    _orderWindowsByMotionAndStartup: function(windows, slots) {
+        let appMonitor = Shell.AppMonitor.get_default();
+        windows.sort(function(w1, w2) {
+            return w2.get_stable_sequence() - w1.get_stable_sequence();
+        });
+        let slotGeometries = slots.map(Lang.bind(this, this._getSlotAbsoluteGeometry));
+        if (windows.length <= POSITIONING_PERMUTATIONS_MAX)
+            return this._orderWindowsPermutations(windows, slots, slotGeometries);
+        else
+            return this._orderWindowsGreedy(windows, slots, slotGeometries);
+    },
+
+    /**
+     * _getSlotRelativeGeometry:
+     * @slot: A layout slot
+     *
+     * Returns: the workspace-relative [x, y, width, height]
+     * of a given window layout slot.
+     */
+    _getSlotRelativeGeometry: function(slot) {
+        let [xCenter, yCenter, fraction] = slot;
+
+        let width = global.screen_width * fraction;
+        let height = global.screen_height * fraction;
+
+        let x = xCenter * global.screen_width - width / 2;
+        let y = yCenter * global.screen_height - height / 2;
+
+        return [x, y, width, height];
+    },
+
+    /**
+     * _getSlotAbsoluteGeometry:
+     * @slot: A layout slot
+     *
+     * Returns: the screen coordiantes [x, y, width, height]
+     * of a given window layout slot.
+     */
+    _getSlotAbsoluteGeometry: function(slot) {
+        let [x, y, width, height] = this._getSlotRelativeGeometry(slot);
+        return [ this.gridX + x, this.gridY + y,
+                 this.scale * width, this.scale * height];
+    },
+
+    /**
+     * _computeWindowRelativeLayout:
+     * @metaWindow: A #MetaWindow
+     * @slot: A layout slot
+     *
+     * Given a window and slot to fit it in, compute its
+     * workspace-relative [x, y, scale] where scale applies
+     * to both X and Y directions.
+     */
+    _computeWindowRelativeLayout: function(metaWindow, slot) {
+        let [xCenter, yCenter, fraction] = slot;
+
+        xCenter = xCenter * global.screen_width;
+        yCenter = yCenter * global.screen_height;
+
+        let rect = new Meta.Rectangle();
+        metaWindow.get_outer_rect(rect);
+
+        let desiredWidth = global.screen_width * fraction;
+        let desiredHeight = global.screen_height * fraction;
+        let scale = Math.min(desiredWidth / rect.width,
+                             desiredHeight / rect.height,
+                             1.0 / this.scale);
+
+        let x = xCenter - 0.5 * scale * rect.width;
+        let y = yCenter - 0.5 * scale * rect.height;
+
+        return [x, y, scale];
+    },
+
     /**
      * positionWindows:
      * @workspaceZooming: If true, then the workspace is moving at the same time and we need to take that into account.
@@ -664,47 +905,25 @@ Workspace.prototype = {
     positionWindows : function(workspaceZooming) {
         let totalVisible = 0;
 
-        for (let i = 1; i < this._windows.length; i++) {
-            let clone = this._windows[i];
+        let visibleWindows = this._getVisibleWindows();
 
-            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows))
-                continue;
+        // Start the animations
+        let slots = this._computeAllWindowSlots(visibleWindows.length);
+        visibleWindows = this._orderWindowsByMotionAndStartup(visibleWindows, slots);
 
-            totalVisible += 1;
-        }
+        for (let i = 0; i < visibleWindows.length; i++) {
+            let slot = slots[i];
+            let metaWindow = visibleWindows[i];
+            let mainIndex = this._lookupIndex(metaWindow);
+            let clone = metaWindow._delegate;
+            let icon = this._windowIcons[mainIndex];
 
-        let previousWindow = this._windows[0];
-        let visibleIndex = 0;
-        for (let i = 1; i < this._windows.length; i++) {
-            let clone = this._windows[i];
-            let icon = this._windowIcons[i];
-
-            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows))
-                continue;
-
-            clone.stackAbove = previousWindow.actor;
-            previousWindow = clone;
-
-            visibleIndex += 1;
-
-            let [xCenter, yCenter, fraction] = this._computeWindowPosition(visibleIndex, totalVisible);
-            xCenter = xCenter * global.screen_width;
-            yCenter = yCenter * global.screen_height;
-
-            // clone.actor.width/height aren't reliably set at this point for
-            // a new window - they're only set when the window contents are
-            // initially updated prior to painting.
-            let cloneRect = new Meta.Rectangle();
-            clone.realWindow.meta_window.get_outer_rect(cloneRect);
-
-            let desiredWidth = global.screen_width * fraction;
-            let desiredHeight = global.screen_height * fraction;
-            let scale = Math.min(desiredWidth / cloneRect.width, desiredHeight / cloneRect.height, 1.0 / this.scale);
+            let [x, y, scale] = this._computeWindowRelativeLayout(metaWindow, slot);
 
             icon.hide();
-            Tweener.addTween(clone.actor, 
-                             { x: xCenter - 0.5 * scale * cloneRect.width,
-                               y: yCenter - 0.5 * scale * cloneRect.height,
+            Tweener.addTween(clone.actor,
+                             { x: x,
+                               y: y,
                                scale_x: scale,
                                scale_y: scale,
                                workspace_relative: workspaceZooming ? this : null,
@@ -717,6 +936,24 @@ Workspace.prototype = {
         }
     },
 
+    syncStacking: function(stackIndices) {
+        let desktopClone = this._windows[0];
+
+        let visibleClones = this._getVisibleClones();
+        visibleClones.sort(function (a, b) { return stackIndices[a.metaWindow.get_stable_sequence()] - stackIndices[b.metaWindow.get_stable_sequence()]; });
+
+        for (let i = 0; i < visibleClones.length; i++) {
+            let clone = visibleClones[i];
+            let metaWindow = clone.metaWindow;
+            if (i == 0) {
+                clone.setStackAbove(desktopClone.actor);
+            } else {
+                let previousClone = visibleClones[i - 1];
+                clone.setStackAbove(previousClone.actor);
+            }
+        }
+    },
+
     _fadeInWindowIcon: function (clone, icon) {
         icon.opacity = 0;
         icon.show();
@@ -1032,11 +1269,7 @@ Workspace.prototype = {
         return clone;
     },
 
-    _computeWindowPosition : function(index, totalWindows) {
-        // ignore this._windows[0], which is the desktop
-        let windowIndex = index - 1;
-        let numberOfWindows = totalWindows;
-
+    _computeWindowSlot : function(windowIndex, numberOfWindows) {
         if (numberOfWindows in POSITIONS)
             return POSITIONS[numberOfWindows][windowIndex];
 
@@ -1053,6 +1286,14 @@ Workspace.prototype = {
         return [xCenter, yCenter, fraction];
     },
 
+    _computeAllWindowSlots: function(totalWindows) {
+        let slots = [];
+        for (let i = 0; i < totalWindows; i++) {
+            slots.push(this._computeWindowSlot(i, totalWindows));
+        }
+        return slots;
+    },
+
     _onCloneSelected : function (clone, time) {
         Main.overview.activateWindow(clone.metaWindow, time);
     },
@@ -1137,10 +1378,12 @@ Workspaces.prototype = {
         // workspaces have been created. This cannot be done first because
         // window movement depends on the Workspaces object being accessible
         // as an Overview member.
-        Main.overview.connect('showing',
-                             Lang.bind(this, function() {
-            for (let w = 0; w < this._workspaces.length; w++)
-                this._workspaces[w].zoomToOverview();
+        this._overviewShowingId =
+            Main.overview.connect('showing',
+                                 Lang.bind(this, function() {
+                this._onRestacked();
+                for (let w = 0; w < this._workspaces.length; w++)
+                    this._workspaces[w].zoomToOverview();
         }));
 
         // Track changes to the number of workspaces
@@ -1150,6 +1393,9 @@ Workspaces.prototype = {
         this._switchWorkspaceNotifyId =
             global.window_manager.connect('switch-workspace',
                                           Lang.bind(this, this._activeWorkspaceChanged));
+        this._restackedNotifyId =
+            global.screen.connect('restacked',
+                                  Lang.bind(this, this._onRestacked));
     },
 
     _lookupWorkspaceForMetaWindow: function (metaWindow) {
@@ -1226,21 +1472,6 @@ Workspaces.prototype = {
         }
     },
 
-    _activateWindowInternal: function (metaWindow, time) {
-        let activeWorkspaceNum = global.screen.get_active_workspace_index();
-        let windowWorkspaceNum = metaWindow.get_workspace().index();
-
-        let clone = this._lookupCloneForMetaWindow (metaWindow);
-        clone.actor.raise_top();
-
-        if (windowWorkspaceNum != activeWorkspaceNum) {
-            let workspace = global.screen.get_workspace_by_index(windowWorkspaceNum);
-            workspace.activate_with_focus(metaWindow, time);
-        } else {
-            metaWindow.activate(time);
-        }
-    },
-
     /**
      * activateWindowFromOverview:
      * @metaWindow: A #MetaWindow
@@ -1253,7 +1484,8 @@ Workspaces.prototype = {
         if (this._windowSelectionAppId != null) {
             this._clearApplicationWindowSelection(false);
         }
-        this._activateWindowInternal(metaWindow, time);
+
+        Main.activateWindow(metaWindow, time);
         Main.overview.hide();
     },
 
@@ -1276,8 +1508,10 @@ Workspaces.prototype = {
         this.actor.destroy();
         this.actor = null;
 
+        Main.overview.disconnect(this._overviewShowingId);
         global.screen.disconnect(this._nWorkspacesNotifyId);
         global.window_manager.disconnect(this._switchWorkspaceNotifyId);
+        global.screen.disconnect(this._restackedNotifyId);
     },
 
     getScale : function() {
@@ -1423,6 +1657,19 @@ Workspaces.prototype = {
         this.actor.add_actor(workspace.actor);
     },
 
+    _onRestacked: function() {
+        let stack = global.get_windows();
+        let stackIndices = {};
+
+        for (let i = 0; i < stack.length; i++) {
+            // Use the stable sequence for an integer to use as a hash key
+            stackIndices[stack[i].get_meta_window().get_stable_sequence()] = i;
+        }
+
+        for (let i = 0; i < this._workspaces.length; i++)
+            this._workspaces[i].syncStacking(stackIndices);
+    },
+
     // Handles a drop onto the (+) button; assumes the new workspace
     // has already been added
     acceptNewWorkspaceDrop : function(source, dropActor, x, y, time) {
