@@ -437,9 +437,10 @@ get_app_for_window_direct (MetaWindow  *window)
 
       if (id != NULL)
         result = shell_app_system_load_from_desktop_file (appsys, id, NULL);
-      else
-        result = create_transient_app_for_window (window);
     }
+  if (result == NULL)
+    result = create_transient_app_for_window (window);
+
   return result;
 }
 
@@ -1038,6 +1039,7 @@ shell_app_monitor_init (ShellAppMonitor *self)
   path = g_build_filename (shell_config_dir, DATA_FILENAME, NULL);
   g_free (shell_config_dir);
   self->configfile = g_file_new_for_path (path);
+  g_free (path);
   restore_from_file (self);
 
   load_initial_windows (self);
