@@ -146,6 +146,63 @@ shell_draw_clock (ClutterCairoTexture *texture,
   cairo_destroy (cr);
 }
 
+void
+shell_draw_box_pointer (ClutterCairoTexture   *texture,
+                        ShellPointerDirection  direction,
+                        ClutterColor          *border_color,
+                        ClutterColor          *background_color)
+{
+  guint width, height;
+  cairo_t *cr;
+
+  clutter_cairo_texture_get_surface_size (texture, &width, &height);
+
+  clutter_cairo_texture_clear (texture);
+  cr = clutter_cairo_texture_create (texture);
+
+  cairo_set_line_width (cr, 1.0);
+
+  clutter_cairo_set_source_color (cr, border_color);
+
+  switch (direction)
+    {
+    case SHELL_POINTER_UP:
+      cairo_move_to (cr, 0, height);
+      cairo_line_to (cr, floor (width * 0.5), 0);
+      cairo_line_to (cr, width, height);
+      break;
+
+    case SHELL_POINTER_DOWN:
+      cairo_move_to (cr, width, 0);
+      cairo_line_to (cr, floor (width * 0.5), height);
+      cairo_line_to (cr, 0, 0);
+      break;
+
+    case SHELL_POINTER_LEFT:
+      cairo_move_to (cr, width, height);
+      cairo_line_to (cr, 0, floor (height * 0.5));
+      cairo_line_to (cr, width, 0);
+      break;
+
+    case SHELL_POINTER_RIGHT:
+      cairo_move_to (cr, 0, 0);
+      cairo_line_to (cr, width, floor (height * 0.5));
+      cairo_line_to (cr, 0, height);
+      break;
+
+    default:
+      g_assert_not_reached();
+    }
+
+  cairo_stroke_preserve (cr);
+
+  clutter_cairo_set_source_color (cr, background_color);
+
+  cairo_fill (cr);
+
+  cairo_destroy (cr);
+}
+
 static void
 hook_paint_red_border (ClutterActor  *actor,
                        gpointer       user_data)
@@ -153,25 +210,20 @@ hook_paint_red_border (ClutterActor  *actor,
   CoglColor color;
   ClutterGeometry geom;
   float width = 2;
-  float x2;
-  float y2;
 
   cogl_color_set_from_4ub (&color, 0xff, 0, 0, 0xc4);
   cogl_set_source_color (&color);
 
   clutter_actor_get_allocation_geometry (actor, &geom);
-  x2 = geom.x + geom.width;
-  y2 = geom.y + geom.height;
 
   /** clockwise order **/
-  cogl_rectangle (geom.x, geom.y,
-                  x2, geom.y + width);
-  cogl_rectangle (x2 - width, geom.y + width,
-                  x2, y2);
-  cogl_rectangle (x2 - width, y2,
-                  geom.x, y2 - width);
-  cogl_rectangle (geom.x + width, y2 - width,
-                  geom.x, geom.y + width);
+  cogl_rectangle (0, 0, geom.width, width);
+  cogl_rectangle (geom.width - width, width,
+                  geom.width, geom.height);
+  cogl_rectangle (0, geom.height,
+                  geom.width - width, geom.height - width);
+  cogl_rectangle (0, geom.height - width,
+                  width, width);
 }
 
 guint
