@@ -13,6 +13,18 @@ ClutterCairoTexture *shell_create_vertical_gradient (ClutterColor *top,
 ClutterCairoTexture *shell_create_horizontal_gradient (ClutterColor *left,
                                                        ClutterColor *right);
 
+typedef enum {
+  SHELL_POINTER_UP,
+  SHELL_POINTER_DOWN,
+  SHELL_POINTER_LEFT,
+  SHELL_POINTER_RIGHT
+} ShellPointerDirection;
+
+void shell_draw_box_pointer (ClutterCairoTexture   *texture,
+                             ShellPointerDirection  direction,
+                             ClutterColor          *border_color,
+                             ClutterColor          *background_color);
+
 void shell_draw_clock (ClutterCairoTexture *texture,
 	               int                  hour,
 	               int                  minute);
